/* (C) 2018 King Kevin <kingkevin@cuvoodoo.info>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* This is the main part of the remote torch controller.
 * It handles all peripherals (IR, serial, torch light)
 */
#include <stdint.h> // Standard Integer Types
#include <stdio.h> // Standard IO facilities
#include <stdlib.h> // General utilities
#include <stdbool.h> // Boolean
#include <avr/io.h> // AVR device-specific IO definitions
#include <avr/interrupt.h> // Interrupts
#include <avr/pgmspace.h> // Program Space Utilities
#include <avr/sleep.h> // sleep utilities
#include <avr/eeprom.h> // EEPROM handling

#include "main.h" // own definitions (e.g. pins)
#ifdef DEBUG
#include "uart.h"
#endif

/* timer prescale = 1 << (TIMER_PRESCALE - 1)
 * use CK/128 prescaler (1.0/(1E6/128))*256 = 32.8 ms, this is the most precise prescaler to measure the longest 28 ms pulse
 */
#define TIMER_PRESCALE 8

/* the timing error allows in % */
#define IR_JITTER_PERCENT 25
/* the timing error allows in us */
#define IR_JITTER_US 200

/* the Nikon shutter trigger mark+space pattern (in us) to detect
 * only use the first part to not have to wait for so long
 * the complete pattern is (in us): 2000, 28000, 400, 1580, 400, 3580, 400, 63200, 2000, 28000, 400, 1580, 400, 3580, 400
 */
static const uint16_t ir_nikon_pattern[] PROGMEM = { 2000, 28000, 400, 1580, 400, 3580, 400 };

// the JITTER tuning is IR receiver specific.
// larger values have high true and false positives
// use longer patterns to match to lower false positives
// this combination has some false positives when shaking the torch, but it fits the purpose (and an operator nearby can re-enable the light)

/* pre-compute tables of the acceptable range
 * this avoid doing long calculations in the ISR, and we have enough free RAM
 * pre-computing it compile time and storing in flash isn't a lot better since reading from flash uses 3 cycles
 */
static uint8_t ir_nikon_pattern_min[COUNT_OF(ir_nikon_pattern)] = {0};
static uint8_t ir_nikon_pattern_max[COUNT_OF(ir_nikon_pattern)] = {0};

// flags, set in the interrupts and handled in the main program
static volatile bool ir_flag = false; // to process a burst

/* torch going out flicker-animation
 * - even index = torch off duration
 * - odd index = torch on duration
 * - unit = T1 overflow (~32.4 ms)
 */
static const uint8_t torch_flicker[] = {3, 3, 3, 15, 3, 3, 10, 5, 3, 3, 3};
static volatile bool torch_tick = false; // on when a tick passed

void led_on(void)
{
	PORTB |= (1 << LED);
}

void led_off(void)
{
	PORTB &= ~(1 << LED);
}

void led_toggle(void)
{
	PINB = (1 << LED);
}

/* switch torch on */
static void torch_on(void)
{
	PORTB |= (1 << TORCH);
}

/* switch torch off */
static void torch_off(void)
{
	PORTB &= ~(1 << TORCH);
}

/* ISR for IR input */
ISR(PCINT0_vect) { // PCI0 Interrupt Vector for PCINT[5:0]
	static volatile uint8_t ir = (1 << IR); // IR signal state, to detect change
	static volatile uint8_t pulse = 0; // pulse index within the burst

	uint8_t time = TCNT1; // save pulse time before it changes

	if (ir_flag) { // don't do pattern matching while the previous match action has not completed
		return;
	}
	TCNT1 = 0; // clear timer

	if (ir == (PINB & (1 << IR))) { // IR signal did not change
		return;
	}
	ir = (PINB & (1 << IR)); // remember change

	if (pulse >= COUNT_OF(ir_nikon_pattern_min) || pulse >= COUNT_OF(ir_nikon_pattern_max)) { // this should not happen
		pulse = 0; // restart pattern matching
	}

	// match pattern
	if (time >= ir_nikon_pattern_min[pulse] && time <= ir_nikon_pattern_max[pulse]) { // pulse matches sequence
#ifdef DEBUG
		led_toggle();
#endif
		pulse++; // go to next pulse
	} else if (ir && time >= ir_nikon_pattern_min[0] && time <= ir_nikon_pattern_max[0]) { // always look for the first burst
		pulse = 1; // header found
	}


	if (pulse == COUNT_OF(ir_nikon_pattern)) { // all pulses have been received
		ir_flag = true; // notify main loop the complete IR sequence has been received
	}
}

/* use timer overflow as torch tick */
ISR(TIMER1_OVF_vect) { // timer 1 overflow
	torch_tick = true; // update time
}

/* initialize I/O pins */
void ioinit(void)
{
	// configure LED
	DDRB |= (1 << LED); // LED is output
	led_off(); // switch LED off

	// configure torch
	DDRB |= (1 << TORCH); // torch is output
	if (eeprom_read_byte(0)) { // torch state is saved in EEPROM
		torch_off(); // switch torch off
	} else {
		torch_on(); // switch torch off
	}

	// configure InfraRed receiver
	DDRB &= ~(1 << IR); // IR (PB0/PCINT0) receiver is input
	PORTB |= (1 << IR); // enable pull-up to reduce noise
	GIFR = (1 << PCIF); // clear interrupt flag
	GIMSK |= (1 << PCIE); // enable interrupt for PCINT[5:0]
	PCMSK |= (1 << IR); // enable interrupt for IR input

	// use timer 1 to measure IR pulses
	TCCR1 = (0 << CTC1) | // use FF as TOP (not OCR1C)
	        (0 << PWM1A) | // use CTC mode (not PWM)
	        (0 << COM1A1) | (0 << COM1A0) | // don't use comparator A
	        (TIMER_PRESCALE << CS10); // use set prescaler
	TIFR = (1 << TOV1); // clear overflow flag
	TIMSK = (1 << TOIE1); // enable overflow interrupt

#ifdef DEBUG
	// use UART as debug ouput
	uart_init();
#endif

	sei(); // enable interrupts
}

/* action to perform when IR trigger is received
 */
static void ir_action()
{
#ifdef DEBUG
	uart_putstr_P(PSTR("IR\n"));
#endif

	if (eeprom_read_byte(0)) {
		torch_on();
		eeprom_update_byte(0, 0x00); // remember we switched on
		for (uint8_t ticks = 0; ticks < 10; ticks++) { // wait a bit before being able to switch off again
			torch_tick = false;
			while (!torch_tick); // wait until tick passed
		}
	} else {
		eeprom_update_byte(0, 0xff); // remember we switched off
		for (uint8_t i = 0; i < COUNT_OF(torch_flicker); i++) {
			if (i % 2) {
				torch_on();
			} else {
				torch_off();
			}
			for (uint8_t ticks = torch_flicker[i]; ticks > 0; ticks--) {
				torch_tick = false;
				while (!torch_tick); // wait until tick passed
			}
		}
		torch_off();
	}
}

int main(void)
{
	ioinit(); // initialize IOs
	set_sleep_mode(SLEEP_MODE_IDLE); // only use idle sleep to keep timer 1 running (and have fast wake-up)

#ifdef DEBUG
	uart_putstr_P(PSTR("\n\ntorch controller using IR remote\n"));
#endif

	const uint16_t tick = 1000000 / (F_CPU/ (((uint16_t)1) << (TIMER_PRESCALE - 1))); // calculate timer tick in us
#ifdef DEBUG
	uart_putstr_P(PSTR("tick: "));
	uart_putuint(tick);
	uart_putstr_P(PSTR(" us\n"));
#endif

	// calculate the min/max IR mark/spaces times from the pattern
	// this is what needs most of the tuning and is IR-demodulator dependant
	for (uint8_t i = 0; i < COUNT_OF(ir_nikon_pattern) && i < COUNT_OF(ir_nikon_pattern_min) && i < COUNT_OF(ir_nikon_pattern_max); i++) {
		uint16_t us = ((pgm_read_word(&ir_nikon_pattern[i]) - IR_JITTER_US) / tick);
		uint16_t percent = ((pgm_read_word(&ir_nikon_pattern[i]) / tick) * (100 - IR_JITTER_PERCENT)) / 100;
		if (us > percent) {
			us = percent;
		}
		 ir_nikon_pattern_min[i] = us;
		ir_nikon_pattern_min[i] = percent;

		us = ((pgm_read_word(&ir_nikon_pattern[i]) + IR_JITTER_US) / tick);
		percent = ((pgm_read_word(&ir_nikon_pattern[i]) / tick) * (100 + IR_JITTER_PERCENT)) / 100;
		if (us < percent) {
			us = percent;
		}
		if (us > UINT8_MAX - 1) {
			ir_nikon_pattern_max[i] = UINT8_MAX;
		} else {
			ir_nikon_pattern_max[i] = us + 1;
		}
	}

	led_on();
	while (true) {
		bool action = false; // to remember if something happened
		if (ir_flag) { // handle IR input
			led_off(); // notify user IT trigger has been detected
			ir_action(); // perform corresponding IR action
			ir_flag = false; // clear flag now to avoid the second half of the pattern to trigger the action again
			led_on(); // IR signal has been processed
			action = true; // we performed an action
		}
		if (action) { // re-run main loop if something has been performed and flags might have changed
			action = false;
		} else { // nothing happend
			sleep_mode(); // got to sleep
		}
	}
	return 0;
}

