/* (C) 2018 King Kevin <kingkevin@cuvoodoo.info>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/* configure the USI to have a UART output */
// standard libraries
#include <stdint.h> // Standard Integer Types
#include <stdbool.h> // Standard Boolean Types
#include <stdio.h> // Standard IO facilities
// AVR libraries
#include <avr/interrupt.h> // Interrupts
#include <avr/io.h> // device-specific IO definitions
#include <avr/pgmspace.h> // Program Space Utilities
// own libraries
#include "main.h" // global definitions
#include "uart.h" // own definitions

// use look up table to get the nibble in LSB first form for sending over USI
static const uint8_t bit_reserse_order_lut[16] PROGMEM = { 0x0, 0x8, 0x4, 0xc, 0x2, 0xa, 0x6, 0xe, 0x1, 0x9, 0x5, 0xd, 0x3, 0xb, 0x7, 0xf };

/* configure USI as UART (only the transmit side) */
void uart_init(void)
{
	cli(); // disable all interrupts why configuring the peripherals
	// configure timer 0 to generate the baud rate
	TCCR0A = (1 << WGM01) | (0 << WGM00); // use CTC mode to restart on compare match
	TCCR0B = 0; // disable clock
	TCNT0 = 0; // reset timer counter
	// set baud rate (closest possible)
	// WARNING this has to match to the prescaler used for the timer
	if (F_CPU % BAUDRATE > BAUDRATE / 2) {
		OCR0A = F_CPU / BAUDRATE - 0;
	} else {
		OCR0A = F_CPU / BAUDRATE - 1;
	}
	OCR0A += 2;// hand adjustment
	// configure USI output
	DDRB |= (1 << PB1); // set USI_DO as output
	PORTB |= (1 << PB1); // set output to idle high
	USIDR = 0xff; // ensure the output is idle high
	USICR = (0 << USISIE) | // don't enable start condition interrupt
	        (0 << USIOIE) | // don't enable counter overflow interrupt
	        (0 << USIWM1) | (1 << USIWM0) | // use three wire mode to use DO
	        (0 << USICS1) | (1 << USICS0) | (0 << USICLK) | // use timer 0 compare match as clock
	        (0 << USITC); // don't toggle clock port pin
	USISR = 0xf0; // reset interrupt flags and counter value
	sei(); // re-enable interrupts
}

void uart_putchar(char c)
{
	if (c == '\n') {
		uart_putchar('\r');
	}
	USIDR = pgm_read_byte(&(bit_reserse_order_lut[c & 0xf])) << 3; // sent first nibble with start bit (USIDR is sent MSB first, but UART data is LSB first)
	USISR = 0xf0; // reset interrupt flags and counter value
	TCNT0 = 0; // reset timer counter
	TCCR0B = (0 << CS02) | (0 << CS01) | (1 << CS00); // start timer (without prescaler to be most precise)
	while ((USISR & 0xf) < 5); // wait until the 5 first bits are sent (start bit plus lower nibble)
	USISR = 0xf0; // reset interrupt flags and counter value
	USIDR = (pgm_read_byte(&(bit_reserse_order_lut[(c >> 4) & 0xf])) << 4) | 0xf; // sent last nibble and stop bit (USIDR is sent MSB first, but UART data is LSB first)
	while ((USISR & 0xf) < 5); // wait until the 5 last bits are sent (higher nobble and stop bit)
	TCCR0B = 0; // stop timer
}

void uart_puthex(uint8_t hex)
{
	uart_putchar('0');
	uart_putchar('x');
	uint8_t nibble = ((hex >> 4) & 0xf);
	if (nibble < 0xa) {
		uart_putchar('0'+nibble);
	} else {
		uart_putchar('a'+nibble-0xa);
	}
	nibble = ((hex >> 0) & 0xf);
	if (nibble < 0xa) {
		uart_putchar('0'+nibble);
	} else {
		uart_putchar('a'+nibble-0xa);
	}
}

void uart_putuint(uint32_t ui)
{
	char digits[10];
	uint8_t digit = 0;
	if (0 == ui) {
		digits[digit++] = '0';
	} else {
		while (ui > 0 && digit < COUNT_OF(digits)) {
			digits[digit++] = '0' + (ui % 10);
			ui /= 10;
		}
	}
	while (digit > 0) {
		digit--;
		uart_putchar(digits[digit]);
	}
}

void uart_putstr(const char* str)
{
	while (*str != '\0') {
		uart_putchar(*str);
		str++;
	}
}

void uart_putstr_P(const char* str)
{
	char c = pgm_read_byte(str);
	while (c != '\0') {
		uart_putchar(c);
		str++;
		c = pgm_read_byte(str);
	}
}

